// $Id:$
//
// WHAT
//
// Daniel Huson, WHEN

#include"options.h"
#include "taxaset.h"
#include "names.h"
#include"time.h"

#include"LEDA/matrix.h"
#include"LEDA/random_source.h"
#include"LEDA/array.h"
#include"LEDA/dictionary.h"
#include"LEDA/list.h"
#include"LEDA/vector.h"

//#include"f2c.h"
//#include"clapack.h"

//extern void singular_vectors(matrix& A, vector& v1, vector& v2);

#define DELTA 0.0000001

// WHAT

double dot_prod2(vector& u, vector& v);
double dot_prod(vector& u, vector& v);
double H_min(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means);
double H_ave(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, int square = 0);
double S_max(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means);
double S_ave(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, int square = 0);
void Klineberg(matrix& A, vector& v, int threshold);
void sqkmeans(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, double t1, int t2,
tOptions& options);
void kmeans(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, double t1);
void intersect(set<string>& A, set<string>& B, set<string>& I);
double JC(set<string>& A, set<string>& B);

int main (const int argc,const char *argv[])
{
    
  // set program options:
  tOptions options;
  options.parse_command_line(argc,argv);
  
  int rows;
  int columns;
  
  matrix em;
  istream &ins = options.ins();
  string instring;
  tNames gene_names;
  dictionary<string, int> gene_names_inv;
  {
    random_source rs(0,1);
    rs.set_seed((int)time(0));
  }
  
  array<vector> true_means(options.num_clusters);
  array<tTaxaSet> true_clusters(options.num_clusters);
  if (options.simulate) {
    matrix data(10*options.num_clusters, options.columns);
    int k = options.num_clusters;
    random_source rs;
  
    if (options.verbose >= 4)
      cout << "generating random means" << endl;

    //generate random true means
    for(int i = 0; i < true_means.size(); i++) {
      vector v(options.columns);
      //double total = 0;
      for(int k = 0; k < v.dim(); k++) {
	rs >> v[k];
	double x;
	rs >> x;
	if (x < 0.5) v[k] = -1*v[k];
	//total += v[k];
      }
      true_means[i] = v.norm();
    }
    
    for(int k = 0; k < options.num_clusters; k++) {
      for(int j = 0; j < options.num_clusters; j++) {
	//cout << dot_prod(true_means[k],true_means[j]) << " ";
      }
      //cout << endl;
    }
    

    if (options.verbose >= 4)
      cout << "generating perturbed data" << endl;

    double noise = options.noise;
    for(int k =0; k < options.num_clusters; k++) {
      for(int i = 0; i < 10; i++) {
	vector v(options.columns);
	double x;
	rs >> x;
	for(int j = 0; j < v.dim(); j++){ 
	  if (x < 0.5) 
	    v[j] = true_means[k][j] + noise;
	  else 
	    v[j] = (-1*true_means[k][j]) - noise;
	}
	data.row(k*10 + i) = v;
	gene_names.insert(k*10 + i, "x");
	true_clusters[k].insert(k*10+i);
      }
    }
    em = data;
    rows = options.num_clusters*10;
    columns = options.columns;

    double obj = 0;
    for(int i = 0; i < true_clusters.size(); i++) {
      tTaxaSet c = true_clusters[i];
      if (c.size() > 0) {
	int g;
	forall(g, c) {
	  obj += dot_prod2(true_means[i], em.row(g));
	}
      }
    }
    cout << "Ideal obj value = " << obj << endl;
  }
  else {
    ins>>rows;
    ins>>columns;
    matrix data(rows, columns);
    em = data;
    
    if (options.verbose >= 4) 
      cout << "reading expression matrix..." << endl;
    
    for(int i = 0; i < rows; i++) {
      ins>>instring;
      gene_names.insert(i, instring); 
      gene_names_inv.insert(instring, i);
      for(int j = 0; j < columns; j++) {
	ins >> instring;
	if (instring != "NA" && instring != "NaN") {
	  char *endptr;
	  double d = strtod(instring.cstring(), &endptr);
	  em(i,j) = d;
	  //ins>>em(i,j);
	  
	}
	else {
	  em(i, j) = 0;
	}
      }
    }
    
    if (options.permute) {
	random_source rs_row(0, em.dim1() - 1);
	random_source rs_col(0, em.dim2() - 1);
	int x1, y1, x2, y2;
	double temp;
	for(int i = 0 ; i < em.dim1()*em.dim2()*5000; i++) {
	    rs_row >> x1;
	    rs_row >> x2;
	    rs_col >> y1;
	    rs_col >> y2;
	    temp = em(x1, y1);
	    em(x1, y1) = em(x2, y2);
	    em(x2, y2) = temp;
	}
    }
  }

  array<tTaxaSet> random_cluster;
  if (options.random_cluster) {
      array<tTaxaSet> rand_cluster(options.num_clusters*2);
      random_source rs(0,options.num_clusters*2 - 1);
      for(int i = 0; i < em.dim1(); i++) {
	  int x;
	  rs >> x;
	  rand_cluster[x].insert(i);
      }
      random_cluster = rand_cluster;
    for(int i = 0; i < random_cluster.size(); i++) {
      cout << random_cluster.size() << " " << random_cluster[i] << endl;
    }
  }
  
  array<int> gene_to_cluster_map(rows);
  if (options.normalize) {
    for(int i = 0; i < rows; i++) {
      em.row(i) = em.row(i).norm();
    }
  }

 array<set<string> > function(rows);
  for(int i = 0; i < rows; i++) function[i].clear();
  if (options.compare_function) {
    //cout << "doing compare function" << endl;
    istream &ins2 = options.ins2();
    string gene, func = "", temp = "";
    while (ins2) {
      ins2 >> gene;
      //cout << gene << endl;
      if (gene.length() > 0) {
	int i = gene_names_inv.access(gene);
	ins2 >> temp; 
	function[i].insert(temp);
	ins2 >> temp;
	while(temp.pos("END") == -1) {
	//while(temp[0] != 'E' && temp[1] != 'N' && temp[2] != 'D') {
	  function[i].insert(temp);
	  ins2 >> temp;
	}
      }
    }
  }
  /*
  array<set<string> > function_cluster(options.num_clusters * 2);  
  dictionary<string, string> function;
  if (options.compare_function) {
    istream &ins2 = options.ins2();
    string gene, func = "", temp = "";
    while (ins2) {
      ins2 >> gene;
      ins2 >> temp; 
      func = "";
      while (temp != "END" && temp != "\0") {
	func = temp + " " + func; 
	ins2 >> temp;
	//cout << temp << endl;
      }
      //cout << temp << endl;
      function.insert(gene, func);
      //cout << func << endl;
    }
  }
  */
  
  cout << "here" << endl;
  if (options.do_data) {
    cout << "doing do_data" << endl;
    int count = 0;
    for(int i = 0; i < rows; i++) {
      for(int j = i+1; j < rows; j++) {
	set<string> A = function[i];
	set<string> B = function[j];
	
	if (A.size() > 0 && B.size() > 0) {
	  double jc = JC(A,B);
	  if (count % 1000 == 0) { cout << count << " are done\n";}
	  count++;
	  vector v1 = em.row(i);
	  vector v2 = em.row(j);
	  double dp;
	  dp = dot_prod2(v1, v2);
	  
	  ostream& outs = options.outs();
	  //outs << jc << " " << dp << " " << gi << " " << gj << endl;
	  outs << jc << " " << dp << endl;
	}
      }
    }
  }
  /*
  if (options.do_data) {
    dic_item it, it2;
    forall_items(it, function) {
      string gi = function.key(it);
      
      forall_items(it2, function) {
	string gj = function.key(it2);
	
	if (gi != gj && gi != "" && gj != "") {
	  string f1 = function.inf(it);
	  string f2 = function.inf(it2);
	  //cout << f1 << endl << f2 << endl << endl;
	  int sp;
	  set<string> A;
	  set<string> B;
	  while((sp = f1.pos(" ")) > -1) {
	    string temp = f1(0,sp);
	    temp = temp.del(temp.length() - 1);
	    f1 = f1.del(0, sp);
	    A.insert(temp);
	  }
	  while((sp = f2.pos(" ")) > -1) {
	    string temp = f2(0,sp);
	    temp = temp.del(temp.length() - 1);
	    f2 = f2.del(0, sp);
	    B.insert(temp);
	  }
	  double jc = JC(A,B);
	  dic_item it3; int i = -1, j = -1;
	  forall_items(it3, gene_names){
	    if (gene_names.inf(it3) == gi) {i = gene_names.key(it3);}
	    if (gene_names.inf(it3) == gj) {j = gene_names.key(it3);}
	    if (i >= 0 && j >= 0) break;
	  }
	  if (i >= 0 && j >=0) {
	    vector v1 = em.row(i);
	    vector v2 = em.row(j);
	    double dp = dot_prod2(v1,v2);
	    ostream& outs = options.outs();
	    outs << jc << " " << dp << endl;
	  }
	}
      }
    }
  }
  */
  
  cout << "start of algorithm" << endl;
  //start of algorithm
  if (options.verbose >= 4) 
    cout <<"random initialization" << endl;
  array<tTaxaSet> cluster_array(options.num_clusters);
  array<vector> means(options.num_clusters);
  
  random_source rs(0, options.num_clusters-1);
  int cl;
  for(int i = 0; i < rows; i++) {
    rs >> cl;
    cluster_array[cl].insert(i);
  }
  if (options.verbose >= 4) 
    cout << "computing initial means..." << endl;
  
  for(int i = 0; i < cluster_array.size(); i++) {
    tTaxaSet s = cluster_array[i];
    if (s.size() > 0) {
      matrix AT(s.size(), columns);
      int g;
      int j = 0;
      forall(g, s) {
	AT.row(j) = em.row(g);
	j++;
      }
      assert(s.size() == j);
      matrix A = AT.trans();
      vector v(columns);
      vector dummy(columns);
      //vector v(s.size());
      //vector dummy(s.size());
      //singular_vectors(A, v, dummy);
      //cout << "here " << endl;
      //cout << v << endl;

      Klineberg(A, v, options.t1);
      means[i] = v.norm();
    }
  }
  
  if (options.verbose >= 4) 
    cout << "calling sqkmeans" << endl;
  
  sqkmeans(em, cluster_array, means, options.t2, options.t1, options);
  
  if (options.simulate) {
    /*
    cout << "comparison of k-svd and random : " << endl;
    for(int i = 0; i < cluster_array.size(); i++) {
      cout << "k-svd clusters = " << cluster_array[i] << endl;
      cout << "true clusters = " << true_clusters[i] << endl;
      tTaxaSet s = cluster_array[i];
      int g;
      forall(g, s)
	cout << dot_prod(em.row(g), true_means[g/10]) << " " ;
      cout << endl;
      forall(g, s)
	cout << dot_prod(em.row(g), true_means[i]) << " " ;
      cout << endl;
      }
    */
  }
  if (options.verbose >= 4) 
    cout << "computing score..." << endl;
  
  double h_ave = H_ave(em, cluster_array, means, 1);
  double h_min = H_min(em, cluster_array, means);
  double s_ave = S_ave(em, cluster_array, means, 1);
  double s_max = S_max(em, cluster_array, means);
  
  cout << "H_AVE = " << h_ave << "  S_AVE = " << s_ave << endl;
  if (options.verbose >= 4) 
    cout << "init new clusters" << endl;

  array<tTaxaSet> cluster_array2(cluster_array.size() * 2);
  array<vector> means2(means.size() * 2);
  vector v(em.dim2());
  for (int k = 0; k < means2.size(); k++) {
      means2[k] = v;
  } 
  
  random_source rs2(0,1);
  for(int i = 0; i < cluster_array.size(); i++) {
      tTaxaSet s = cluster_array[i];
      
    if (s.size() > 0) {
      array<tTaxaSet> temp_cluster_array(2);
      array<vector> temp_means(2);
      vector v(em.dim2());
      temp_means[0] = temp_means[1] = v;
      
      int g;
      double a = 0, b = 0;
      forall(g, s) {
	int x = 0;
	if (dot_prod(em.row(g), means[i]) > 0)  x = 1;
	else x = 0;
	
	//int x = rs2();
	if (x == 0) a++; else b++;
	temp_cluster_array[x].insert(g);
	temp_means[x] = temp_means[x] + em.row(g);
      }
      assert(a > 0 || b > 0);
      if (a == 0) a = 1; if (b == 0) b = 1;
      temp_means[0] = temp_means[0]*1/a;
      temp_means[1] = temp_means[1]*1/b;
      temp_means[0] = temp_means[0].norm();
      temp_means[1] = temp_means[1].norm();
      
      kmeans(em, temp_cluster_array, temp_means, options.t2);

      cluster_array2[i*2 + 0] = temp_cluster_array[0];
      cluster_array2[i*2 + 1] = temp_cluster_array[1];    
      means2[i*2 + 0] = temp_means[0];
      means2[i*2 + 1] = temp_means[1];    
      
      if (options.random_cluster) {
	int g;
	tTaxaSet S = random_cluster[i*2];
	vector w(em.dim2());
	forall (g, S) {
	  vector v = em.row(g);
	  w  = w + v;
	}
	w = w * 1/(double)S.size();
	means2[i*2] = w;

	S = random_cluster[i*2 + 1];
	forall (g, S) {
	  vector v = em.row(g);
	  w  = w + v;
	}
	w = w * 1/(double)S.size();
	means2[i*2 + 1] = w;

	cluster_array2[i*2] = random_cluster[i*2];
	cluster_array2[i*2 + 1] = random_cluster[i*2 + 1];
	temp_cluster_array[0] = random_cluster[i*2];
	temp_cluster_array[1] = random_cluster[i*2 + 1];
	temp_means[0] = means2[i*2];
	temp_means[1] = means2[i*2 + 1];
      }
      
      s = temp_cluster_array[0];
      forall(g, s) gene_to_cluster_map[g] = i*2 + 0;
      if (s.empty()) {
	for(int k = 0; k < temp_means[0].dim(); k++) 
	  temp_means[0][k] = 0;
      }
      s = temp_cluster_array[1];
      forall(g, s) gene_to_cluster_map[g] = i*2 + 1;
      if (s.empty()) {
	for(int k = 0; k < temp_means[1].dim(); k++) 
	  temp_means[1][k] = 0;
      }

      cout << "# BEGIN DIAM" << endl;
      cout << "CLUSTER " << i*2 + 0 << endl;
      if (cluster_array2[i*2].size() > 0) 
	{gene_names.print(cluster_array2[i*2]); cout << endl;}
      else 
	cout << "EMPTY" << endl;
      cout << temp_means[0] << endl;
      cout << "CLUSTER " << i*2 + 1 << endl;
      if (cluster_array2[i*2 + 1].size() > 0) 
	{gene_names.print(cluster_array2[i*2 + 1]); cout << endl;}
      else 
	cout << "EMPTY" << endl;
      cout << temp_means[1] << endl;
      cout << "h_ave = " << H_ave(em, temp_cluster_array, temp_means) << "  "
	   << "s_ave = " << S_ave(em, temp_cluster_array, temp_means) << endl;


      if (options.compare_function) {
	istream &ins2 = options.ins2();
	string gene, func = "", temp = "";
	while (ins2) {
	  ins2 >> gene;
	  //cout << gene << endl;
	  if (gene.length() > 0) {
	    int i = gene_names_inv.access(gene);
	    ins2 >> temp; 
	    function[i].insert(temp);
	    ins2 >> temp;
	    while(temp.pos("END") == -1) {
	      //while(temp[0] != 'E' && temp[1] != 'N' && temp[2] != 'D') {
	      function[i].insert(temp);
	      ins2 >> temp;
	    }
	  }
	}
      }
      
      if (options.compare_function) {
	set<string> A, AG, B, BG;A.clear();B.clear();
	tTaxaSet s;
	tTaxaSet t;
	if (options.random_cluster) {
	  s = random_cluster[i*2];
	  t = random_cluster[i*2 + 1];
	}
	else {
	  s = cluster_array2[i*2];
	  t = cluster_array2[i*2 + 1];
	}
	int g, f;
	int total = 0;
	double total_jc = 0;
	forall(g, s) {
	    forall(f, t) {

	      set<string> A = function[g];
	      set<string> B = function[f];
	      if (A.size() > 0 && B.size() > 0) {
		double jc = JC(A,B);
		total_jc += JC(A,B);
		total++;
	      }
	      B.clear();
	    }
	    A.clear();
	}
	cout << "Diametric JC = " << total_jc/(double)total << endl;
	cout << "total pairs = " << s.size()*t.size() << " function pairs = " << total << endl;
      	
	A.clear();B.clear();
	total = 0;
	total_jc = 0;
	
	forall(g, s) {
	    forall(f, s) {
	      if (f!=g) {
		set<string> A = function[g];
		set<string> B = function[f];
		if (A.size() > 0 && B.size() > 0) {
		  double jc = JC(A,B);
		  total_jc += JC(A,B);
		  total++;
		}
	      }
	      B.clear();
	    }
	    A.clear();
	}
	
	cout << "Single JC for cluster " << i*2 << " = " << total_jc/(double)total << endl;
	cout << "total pairs = " << s.size()*s.size() << " function pairs = " << total << endl;
	
	total = 0;
	total_jc = 0;
	A.clear();B.clear();
	forall(g, t) {
	  forall(f, t) {
	      if (f!=g) {
		set<string> A = function[g];
		set<string> B = function[f];
		if (A.size() > 0 && B.size() > 0) {
		  double jc = JC(A,B);
		  total_jc += JC(A,B);
		  total++;
		}
	      }
	      B.clear();
	  }
	  A.clear();	
	}
	
	cout << "Single JC for cluster " << i*2 +1<< " = " << total_jc/(double)total << endl;
	cout << "total pairs = " << t.size()*t.size() << " function pairs = " << total << endl;
      }
      cout << "END DIAM" << endl;
    }
  }
  
  if (options.outfile != "stdout") {
    ostream& outs = options.outs();
    ostream& outs2 = options.outs2();
    outs << rows << endl;
    for(int i = 0; i < gene_to_cluster_map.size(); i++) {
      outs << gene_to_cluster_map[i] << "\tC" << gene_to_cluster_map[i] << "\t\n";
      double dp = dot_prod(em.row(i), means2[gene_to_cluster_map[i]]);
      outs2 << dp << endl;
    }
  }
  
  if (options.verbose >= 4) 
    cout << "computing score..." << endl;
  
  h_ave = H_ave(em, cluster_array2, means2);
  h_min = H_min(em, cluster_array2, means2);
  s_ave = S_ave(em, cluster_array2, means2);
  s_max = S_max(em, cluster_array2, means2);
  cout << "H_AVE = " << h_ave << "  H_MIN = " << h_min <<
    "  S_AVE = " << s_ave << "  S_MAX = " << s_max << endl;
  
  if (options.simulate) {
    cout << "clusters after kmeans are : " << endl;
      for(int i = 0; i < cluster_array2.size(); i++) {
	//gene_names.print(cluster_array2[i]);
	cout << cluster_array2[i] << endl;
	tTaxaSet s = cluster_array2[i];
	int g;
	forall (g, s) {
	  //cout << dot_prod(em.row(g), true_means[g/10]) << " ";
	  //cout << means2[i] << endl;
	}
	//cout << endl << endl;  
      }
  }
  
} 



double JC(set<string>& A, set<string>& B) {
  set<string> I;
  intersect(A,B,I);
  double jc = (double)(I.size())/(double)(A.size() + B.size() - I.size());
  return jc;
}

void intersect(set<string>& A, set<string>& B, set<string>& I) {
  string a,b;
  forall(a, A) {
    if (B.member(a))
      I.insert(a);
  }
}
	  

void Klineberg(matrix& A, vector& v, int threshold){ 
  vector a(A.dim2());
  vector h(A.dim1());
  vector a_(A.dim2());
  vector h_(A.dim1());

  for (int i = 0; i < h.dim(); i++) {
    h[i] = 1;
  }

  for (int i = 0; i < a.dim(); i++) {
    a[i] = 1;
  }

  for (int i = 0; i < threshold; i++){
    a_ = A.trans()*h;
    h_ = A*a;
    h = h_.norm();
    a = a_.norm();
  }
  v = h;
}

void sqkmeans(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, double threshold, int t1,
tOptions& options){
  

  array<int> empty(cluster_array.size());
  for (int i = 0; i < empty.size(); i++) {
    empty[i] = 0;
  }
  assert(cluster_array.size() == means.size());
  
  double obj = 0;
  for(int i = 0; i < cluster_array.size(); i++) {
    tTaxaSet c = cluster_array[i];
    if (c.size() > 0) {
      int g;
      forall(g, c) {
	obj += dot_prod2(means[i], em.row(g));
      }
    }
  }
  double prev_obj = obj;
  
  //  for (int k = 0; k < threshold; k++) {
  while (1) {
    if (options.verbose >= 4) 
      cout << "reset clusters for next round" << endl;
  
    for(int i = 0; i < cluster_array.size(); i++) {
      cluster_array[i].clear();
    }
    
    for(int i = 0; i < em.dim1(); i++) {
      vector v = em.row(i);
      double best_score = -2;
      int closest_mean = -1;
      
      for(int j = 0; j < means.size(); j++) {
	if (!empty[j]) {
	  vector w = means[j];
	  assert(w.length() <= 1 + DELTA);
	  double dp = dot_prod2(w, v);
	  if (dp > best_score) {
	    best_score = dp;
	    closest_mean = j;
	  }
	}
      }
      assert(closest_mean > -1);
      cluster_array[closest_mean].insert(i);
    }
    
    if (options.verbose >= 4) 
      cout << "update the means" << endl;
    
    for(int i = 0; i < cluster_array.size(); i++) {
      tTaxaSet s = cluster_array[i];
      if (s.size() > 0) {
	matrix AT(s.size(), em.dim2());
	int g;
	int j = 0;
	forall(g, s) {
	  AT.row(j) = em.row(g);
	  j++;
	}
	assert(s.size() == j);
	matrix A = AT.trans();
	vector v(em.dim2());
	Klineberg(A, v, t1);
	means[i] = v.norm();
      }
      else 
	empty[i] = 1;
    }
    
    double obj = 0;
    for(int i = 0; i < cluster_array.size(); i++) {
      tTaxaSet c = cluster_array[i];
      if (c.size() > 0) {
	int g;
	forall(g, c) {
	  obj += dot_prod2(means[i], em.row(g));
	}
      }
    }
    if (options.verbose >= 4) {
      cout << prev_obj << " " << obj << endl;
    }
    assert(obj + DELTA >= prev_obj);
    if (obj - prev_obj < threshold) 
      break;
    prev_obj = obj;
    
  }
}

void kmeans(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, double threshold){

  tTaxaSet S;
  for(int i = 0; i < cluster_array.size(); i++) {
    S += cluster_array[i];
  }
  
  array<int> empty(cluster_array.size());
  for (int i = 0; i < empty.size(); i++) {
    empty[i] = 0;
  }

  double obj = 0;
  for(int i = 0; i < cluster_array.size(); i++) {
    tTaxaSet c = cluster_array[i];
    if (c.size() > 0) {
      int g;
      forall(g, c) {
	obj += dot_prod(means[i], em.row(g));
      }
    }
  }
  double prev_obj = obj;

  //for (int k = 0; k < threshold; k++) {
  while (1) {
    //reset clusters for next round
    for(int i = 0; i < cluster_array.size(); i++) {
      cluster_array[i].clear();
    }
    
    int g;
    forall (g, S) {
      vector v = em.row(g);
      double best_score = -1000;
      int closest_mean = -1;
      
      for(int j = 0; j < means.size(); j++) {
	if (!empty[j]) {
	  vector w = means[j];
	  double dp = dot_prod(w, v);
	  if (dp > best_score) {
	    best_score = dp;
	    closest_mean = j;
	  }
	}
      }
      cluster_array[closest_mean].insert(g);
    }
    
    //update the means
    for(int i = 0; i < cluster_array.size(); i++) {
      tTaxaSet s = cluster_array[i];
      if (s.size() > 0) {
	vector v(em.dim2());
	taxaset_item it;
	forall_items(it, s) {
	  v = v + em.row(s[it]);
	}
	double inv = 1/(double)(s.size());
	v = v * inv;
	means[i] = v.norm();
	empty[i] = 0;
      }
      else 
	empty[i] = 1;
    }
  
    double obj = 0;
    for(int i = 0; i < cluster_array.size(); i++) {
      tTaxaSet c = cluster_array[i];
      if (c.size() > 0) {
	int g;
	forall(g, c) {
	  obj += dot_prod(means[i], em.row(g));
	}
      }
    }
    cout << prev_obj << " " << obj << endl;
    assert(obj + DELTA>= prev_obj);
    if (obj - prev_obj <= threshold) 
      break;
    prev_obj = obj;    
  }
}


double S_ave(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, int square = 0) {
  
  double s_ave = 0;
  double total = 0;
  for(int i = 0; i < cluster_array.size(); i++) {
    for(int j = i+1; j < cluster_array.size(); j++) {

      tTaxaSet s1 = cluster_array[i];
      tTaxaSet s2 = cluster_array[j];
      
      if (s1.size() > 0 && s2.size() > 0) {
	if (square)
	  s_ave += dot_prod2(means[i], means[j])*s1.size()*s2.size();
	else 
	  s_ave += dot_prod(means[i], means[j])*s1.size()*s2.size();	  

	total += s1.size()*s2.size();
      }
    }
  }
  return s_ave/total;
}

double S_max(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means) {
  
  double s_max = 0;

  for(int i = 0; i < cluster_array.size(); i++) {
    for(int j = i+1; j < cluster_array.size(); j++) {

      tTaxaSet s1 = cluster_array[i];
      tTaxaSet s2 = cluster_array[j];
      
      if (s1.size() > 0 && s2.size() > 0) {
	double dp = dot_prod(means[i], means[j]);
	if (dp > s_max) s_max = dp;
      }
    }
  }
  return s_max;
}

double H_ave(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means, int square = 0) {
  
  tTaxaSet S;
  for(int i = 0; i < cluster_array.size(); i++) {
    S += cluster_array[i];
  }
  
  double h_ave = 0;
  for(int i = 0; i < cluster_array.size(); i++) {
    tTaxaSet s = cluster_array[i];
    if (s.size() > 0) {
      int j;
      if (square)
	forall(j, s) h_ave += dot_prod2(em.row(j), means[i]);
      else
	forall(j, s) h_ave += dot_prod(em.row(j), means[i]);
    }
  }
  return h_ave/(double)S.size();
}

double H_min(matrix& em, array<tTaxaSet>& cluster_array, array<vector>& means) {
  
  double h_min = 1000000;
  for(int i = 0; i < cluster_array.size(); i++) {
    tTaxaSet s = cluster_array[i];
    if (s.size() > 0) {
      int j;
      forall(j, s) {
	double dp = dot_prod(em.row(j),means[i]);
	if (dp < h_min) h_min = dp;
      }
    }
  }
  return h_min;
}

double dot_prod2(vector& u, vector& v){ 
  
  double s = 0;
  assert(u.dim() == v.dim());
  for(int i = 0;i < u.dim();i++) {
    s += u[i]*v[i];
  }
  return s*s;
}

double dot_prod(vector& u, vector& v){ 
  
  double s = 0;
  assert(u.dim() == v.dim());
  for(int i = 0;i < u.dim();i++) {
    s += u[i]*v[i];
  }
  return s;
}

// EOF
