
// $Id: options.c,v 1.3 1998/01/26 18:54:16 huson Exp $
//
// Program options
//
// Daniel Huson, 10.97

#include<stdio.h>
#include<LEDA/basic.h>
#include<LEDA/string.h>
#include<LEDA/stream.h>
#include<LEDA/list.h>

#include"basic.h"
#include"options.h"
#include"version.h"

// set the default options:

tOptions::tOptions (void)
{
	insp=0;
	insp2=0;
	outsp=0;
	outsp2=0;
	outsp3=0;

	infile=string("stdin");
	outfile=string("stdout");

	verbose=0;
	num_clusters = 5;
	t1 = 20;
	t2 = 0.001;
	normalize = 0;

	simulate = 0;
	columns = 10;
	noise = 0.001;
	
	compare_function = 0;
	random_cluster = 0;
	permute = 0;

	do_data = 0;
	init_clusters_out = 0;
	string init_clusters = "";
}
// destructor:

tOptions::~tOptions ()
{
	if(insp)
		delete insp;
	if(insp2)
		delete insp2;
	if(outsp)
		delete outsp;
	if(outsp2)
		delete outsp2;
	if(outsp3)
		delete outsp3;
}

// Parse the command line options:

void tOptions::parse_command_line (int argc,const char *argv[])
{
	string name;

	if(getoption(argc,argv,"--help"))
		usage();

	// Main switchs:

	// Input-output:
	if(getoption(argc,argv,"-i",name))
	{
		infile=name;
		insp=new file_istream(infile);
		if(! *insp)
			bye("Can't open file to read: %s",infile.cstring());
	}
	if(getoption(argc,argv,"-i2",name))
	{
		infile=name;
		insp2=new file_istream(infile);
		if(! *insp2)
			bye("Can't open file to read: %s",infile.cstring());
	}
	if(getoption(argc,argv,"-o",name))
	{
		outfile=name;
		outsp=new file_ostream(outfile);
		if(! outsp)
			bye("Can't open file to write: %s",outfile.cstring());
	}

	if(getoption(argc,argv,"-o2",name))
	{
	  outsp2=new file_ostream(name);
	  if(! outsp2)
	    bye("Can't open file to write: %s",outfile.cstring());
	}

	if(getoption(argc,argv,"-o3",name))
	{
	  outsp3=new file_ostream(name);
	  if(! outsp3)
	    bye("Can't open file to write: %s",outfile.cstring());
	}

	getoption(argc, argv, "-num_clusters", num_clusters);
	getoption(argc, argv, "-init_clusters", init_clusters);
	getoption(argc, argv, "-t1", t1);
	getoption(argc, argv, "-t2", t2);
	if (getoption(argc, argv, "-normalize")) {
	  normalize = 1;
	}
	if (getoption(argc, argv, "-simulate")) {
	  simulate = 1;
	}
	if (getoption(argc, argv, "-compare_function")) {
	  compare_function = 1;
	}
	if (getoption(argc, argv, "-random_cluster")) {
	  random_cluster = 1;
	}
	if (getoption(argc, argv, "-permute")) {
	  permute = 1;
	}
	if (getoption(argc, argv, "-do_data")) {
	  do_data = 1;
	}
	if (getoption(argc, argv, "-init_clusters_out")) {
	  init_clusters_out = 1;
	}

	getoption(argc, argv, "-noise", noise);
	getoption(argc, argv, "-columns", columns);
	

	if(getoption(argc,argv,"-v"))
		verbose=1;
	if(getoption(argc,argv,"-v1"))
		verbose=1;
	if(getoption(argc,argv,"-v2"))
		verbose=2;
	if(getoption(argc,argv,"-v3"))
		verbose=3;
	if(getoption(argc,argv,"-v4"))
		verbose=4;

	// have we got any illegal options:
	list<string> remaining;
	getoption(argc,argv,remaining);
	if(remaining.size()>0)
	{
		cerr<<"Error: Unknown option(s): ";
		remaining.print(cerr);
		cerr<<endl;
		usage();
	}

	messages(verbose);
	if(verbose)
        cerr<<"# "<<version()<<endl;
	if(verbose>1)
		report();
}	

// Report the set options:

void tOptions::report (void) const
{
	cerr
	<<"# Input: "<<infile<<endl
	<<"# Output: "<<outfile<<endl;

}

// returns the input stream:

istream &tOptions::ins (void) const
{
	if(insp)
		return *insp;
	else
		return cin;
}

istream &tOptions::ins2 (void) const
{
	if(insp2)
		return *insp2;
	else
		return cin;
}

// returns the output stream:

ostream &tOptions::outs (void) const
{
	if(outsp)
		return *outsp;
	else
		return cout;
}

ostream &tOptions::outs2 (void) const
{
	if(outsp2)
		return *outsp2;
	else
		return cout;
}

ostream &tOptions::outs3 (void) const
{
	if(outsp3)
		return *outsp3;
	else
		return cout;
}

// Reports the usage and then quits:

void tOptions::usage (void) const
{
	cerr<<version()<<endl
	<<"Usage: program [options]\n"
	<<"Main options:\n"
	<<"Input-output:\n"
	<<"\t-i name    - Read input from named file, not from stdin\n"
	<<"\t-o name    - Write output to named file, not to stdout\n"
	<<"Additional:\n"
#if 0
	<<"\t-v         - verbose mode 1\n"
	<<"\t-v2        - verbose mode 2\n"
	<<"\t-v3        - verbose mode 3\n"
	<<"\t-v4        - verbose mode 4\n"
#endif
	<<"\t--help     - Show this message and then quit\n";
	exit(0);
}
	// EOF
